@@ -7,7 +7,7 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2000-2007 Julian Seward 
+   Copyright (C) 2000-2009 Julian Seward 
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -30,12 +30,15 @@
 
 #include "pub_core_basics.h"
 #include "pub_core_debuglog.h"
-#include "pub_core_execontext.h"    // self
 #include "pub_core_libcassert.h"
 #include "pub_core_libcprint.h"     // For VG_(message)()
 #include "pub_core_mallocfree.h"
 #include "pub_core_options.h"
 #include "pub_core_stacktrace.h"
+#include "pub_core_machine.h"       // VG_(get_IP)
+#include "pub_core_vki.h"           // To keep pub_core_threadstate.h happy
+#include "pub_core_threadstate.h"   // VG_(is_valid_tid)
+#include "pub_core_execontext.h"    // self
 
 /*------------------------------------------------------------*/
 /*--- Low-level ExeContext storage.                        ---*/
@@ -72,10 +75,16 @@ static SizeT ec_primes[N_EC_PRIMES] = {
 
 struct _ExeContext {
    struct _ExeContext* chain;
-   UInt   n_ips;
+   /* A 32-bit unsigned integer that uniquely identifies this
+      ExeContext.  Memcheck uses these for origin tracking.  Values
+      must be nonzero (else Memcheck's origin tracking is hosed), must
+      be a multiple of four, and must be unique.  Hence they start at
+      4. */
+   UInt ecu;
    /* Variable-length array.  The size is 'n_ips'; at
       least 1, at most VG_DEEPEST_BACKTRACE.  [0] is the current IP,
       [1] is its caller, [2] is the caller of [1], etc. */
+   UInt n_ips;
    Addr ips[0];
 };
 
@@ -85,6 +94,9 @@ static ExeContext** ec_htab; /* array [ec_htab_size] of ExeContext* */
 static SizeT        ec_htab_size;     /* one of the values in ec_primes */
 static SizeT        ec_htab_size_idx; /* 0 .. N_EC_PRIMES-1 */
 
+/* ECU serial number */
+static UInt ec_next_ecu = 4; /* We must never issue zero */
+
 
 /* Stats only: the number of times the system was searched to locate a
    context. */
@@ -112,7 +124,7 @@ static void init_ExeContext_storage ( void )
 {
    Int i;
    static Bool init_done = False;
-   if (init_done)
+   if (LIKELY(init_done))
       return;
    ec_searchreqs = 0;
    ec_searchcmps = 0;
@@ -123,7 +135,7 @@ static void init_ExeContext_storage ( void )
 
    ec_htab_size_idx = 0;
    ec_htab_size = ec_primes[ec_htab_size_idx];
-   ec_htab = VG_(arena_malloc)(VG_AR_EXECTXT, 
+   ec_htab = VG_(arena_malloc)(VG_AR_EXECTXT, "execontext.iEs1",
                                sizeof(ExeContext*) * ec_htab_size);
    for (i = 0; i < ec_htab_size; i++)
       ec_htab[i] = NULL;
@@ -137,18 +149,18 @@ void VG_(print_ExeContext_stats) ( void )
 {
    init_ExeContext_storage();
    VG_(message)(Vg_DebugMsg, 
-      "   exectx: %,lu lists, %,llu contexts (avg %,llu per list)",
-      ec_htab_size, ec_totstored, ec_totstored / ec_htab_size
+      "   exectx: %'lu lists, %'llu contexts (avg %'llu per list)\n",
+      ec_htab_size, ec_totstored, ec_totstored / (ULong)ec_htab_size
    );
    VG_(message)(Vg_DebugMsg, 
-      "   exectx: %,llu searches, %,llu full compares (%,llu per 1000)",
+      "   exectx: %'llu searches, %'llu full compares (%'llu per 1000)\n",
       ec_searchreqs, ec_searchcmps, 
       ec_searchreqs == 0 
-         ? 0L 
-         : ( (ec_searchcmps * 1000) / ec_searchreqs ) 
+         ? 0ULL 
+         : ( (ec_searchcmps * 1000ULL) / ec_searchreqs ) 
    );
    VG_(message)(Vg_DebugMsg, 
-      "   exectx: %,llu cmp2, %,llu cmp4, %,llu cmpAll",
+      "   exectx: %'llu cmp2, %'llu cmp4, %'llu cmpAll\n",
       ec_cmp2s, ec_cmp4s, ec_cmpAlls 
    );
 }
@@ -161,7 +173,7 @@ void VG_(pp_ExeContext) ( ExeContext* ec )
 }
 
 
-/* Compare two ExeContexts, comparing all callers. */
+/* Compare two ExeContexts.  Number of callers considered depends on res. */
 Bool VG_(eq_ExeContext) ( VgRes res, ExeContext* e1, ExeContext* e2 )
 {
    Int i;
@@ -248,7 +260,7 @@ static void resize_ec_htab ( void )
       return; /* out of primes - can't resize further */
 
    new_size = ec_primes[ec_htab_size_idx + 1];
-   new_ec_htab = VG_(arena_malloc)(VG_AR_EXECTXT,
+   new_ec_htab = VG_(arena_malloc)(VG_AR_EXECTXT, "execontext.reh1",
                                    sizeof(ExeContext*) * new_size);
 
    VG_(debugLog)(
@@ -277,28 +289,54 @@ static void resize_ec_htab ( void )
    ec_htab_size_idx++;
 }
 
-ExeContext* VG_(record_ExeContext) ( ThreadId tid, Word first_ip_delta )
+/* Do the first part of getting a stack trace: actually unwind the
+   stack, and hand the results off to the duplicate-trace-finder
+   (_wrk2). */
+static ExeContext* record_ExeContext_wrk2 ( Addr* ips, UInt n_ips ); /*fwds*/
+static ExeContext* record_ExeContext_wrk ( ThreadId tid, Word first_ip_delta,
+                                           Bool first_ip_only )
 {
-   Int         i;
-   Addr        ips[VG_DEEPEST_BACKTRACE];
-   Bool        same;
-   UWord       hash;
-   ExeContext* new_ec;
-   ExeContext* list;
-   UInt        n_ips;
-   ExeContext  *prev2, *prev;
+   Addr ips[VG_DEEPEST_BACKTRACE];
+   UInt n_ips;
 
-   static UInt ctr = 0;
+   init_ExeContext_storage();
 
    vg_assert(sizeof(void*) == sizeof(UWord));
    vg_assert(sizeof(void*) == sizeof(Addr));
 
-   init_ExeContext_storage();
+   vg_assert(VG_(is_valid_tid)(tid));
+
    vg_assert(VG_(clo_backtrace_size) >= 1 &&
              VG_(clo_backtrace_size) <= VG_DEEPEST_BACKTRACE);
 
-   n_ips = VG_(get_StackTrace)( tid, ips, VG_(clo_backtrace_size),
-                                first_ip_delta );
+   if (first_ip_only) {
+      n_ips = 1;
+      ips[0] = VG_(get_IP)(tid);
+   } else {
+      n_ips = VG_(get_StackTrace)( tid, ips, VG_(clo_backtrace_size),
+                                   NULL/*array to dump SP values in*/,
+                                   NULL/*array to dump FP values in*/,
+                                   first_ip_delta );
+   }
+
+   return record_ExeContext_wrk2 ( ips, n_ips );
+}
+
+/* Do the second part of getting a stack trace: ips[0 .. n_ips-1]
+   holds a proposed trace.  Find or allocate a suitable ExeContext.
+   Note that callers must have done init_ExeContext_storage() before
+   getting to this point. */
+static ExeContext* record_ExeContext_wrk2 ( Addr* ips, UInt n_ips )
+{
+   Int         i;
+   Bool        same;
+   UWord       hash;
+   ExeContext* new_ec;
+   ExeContext* list;
+   ExeContext  *prev2, *prev;
+
+   static UInt ctr = 0;
+
    tl_assert(n_ips >= 1 && n_ips <= VG_(clo_backtrace_size));
 
    /* Now figure out if we've seen this one before.  First hash it so
@@ -357,13 +395,22 @@ ExeContext* VG_(record_ExeContext) ( ThreadId tid, Word first_ip_delta )
    /* Bummer.  We have to allocate a new context record. */
    ec_totstored++;
 
-   new_ec = VG_(arena_malloc)( VG_AR_EXECTXT, 
+   new_ec = VG_(arena_malloc)( VG_AR_EXECTXT, "execontext.rEw2.2",
                                sizeof(struct _ExeContext) 
                                + n_ips * sizeof(Addr) );
 
    for (i = 0; i < n_ips; i++)
       new_ec->ips[i] = ips[i];
 
+   vg_assert(VG_(is_plausible_ECU)(ec_next_ecu));
+   new_ec->ecu = ec_next_ecu;
+   ec_next_ecu += 4;
+   if (ec_next_ecu == 0) {
+      /* Urr.  Now we're hosed; we emitted 2^30 ExeContexts already
+         and have run out of numbers.  Not sure what to do. */
+      VG_(core_panic)("m_execontext: more than 2^30 ExeContexts created");
+   }
+
    new_ec->n_ips = n_ips;
    new_ec->chain = ec_htab[hash];
    ec_htab[hash] = new_ec;
@@ -378,11 +425,55 @@ ExeContext* VG_(record_ExeContext) ( ThreadId tid, Word first_ip_delta )
    return new_ec;
 }
 
-StackTrace VG_(extract_StackTrace) ( ExeContext* e )
-{                                  
+ExeContext* VG_(record_ExeContext)( ThreadId tid, Word first_ip_delta ) {
+   return record_ExeContext_wrk( tid, first_ip_delta,
+                                      False/*!first_ip_only*/ );
+}
+
+ExeContext* VG_(record_depth_1_ExeContext)( ThreadId tid ) {
+   return record_ExeContext_wrk( tid, 0/*first_ip_delta*/,
+                                      True/*first_ip_only*/ );
+}
+
+ExeContext* VG_(make_depth_1_ExeContext_from_Addr)( Addr a ) {
+   init_ExeContext_storage();
+   return record_ExeContext_wrk2( &a, 1 );
+}
+
+StackTrace VG_(get_ExeContext_StackTrace) ( ExeContext* e ) {
    return e->ips;
 }  
 
+UInt VG_(get_ECU_from_ExeContext)( ExeContext* e ) {
+   vg_assert(VG_(is_plausible_ECU)(e->ecu));
+   return e->ecu;
+}
+
+Int VG_(get_ExeContext_n_ips)( ExeContext* e ) {
+   vg_assert(e->n_ips >= 1);
+   return e->n_ips;
+}
+
+ExeContext* VG_(get_ExeContext_from_ECU)( UInt ecu )
+{
+   UWord i;
+   ExeContext* ec;
+   vg_assert(VG_(is_plausible_ECU)(ecu));
+   vg_assert(ec_htab_size > 0);
+   for (i = 0; i < ec_htab_size; i++) {
+      for (ec = ec_htab[i]; ec; ec = ec->chain) {
+         if (ec->ecu == ecu)
+            return ec;
+      }
+   }
+   return NULL;
+}
+
+ExeContext* VG_(make_ExeContext_from_StackTrace)( Addr* ips, UInt n_ips )
+{
+   return record_ExeContext_wrk2(ips, n_ips);
+}
+
 /*--------------------------------------------------------------------*/
 /*--- end                                           m_execontext.c ---*/
 /*--------------------------------------------------------------------*/
