@@ -7,9 +7,9 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2000-2006 Julian Seward 
+   Copyright (C) 2000-2009 Julian Seward 
       jseward@acm.org
-   Copyright (C) 2003-2006 Jeremy Fitzhardinge
+   Copyright (C) 2003-2009 Jeremy Fitzhardinge
       jeremy@goop.org
 
    This program is free software; you can redistribute it and/or
@@ -36,6 +36,7 @@
 #include "pub_core_libcbase.h"
 #include "pub_core_libcassert.h"
 #include "pub_core_libcprint.h"
+#include "pub_core_seqmatch.h"
 #include "pub_core_mallocfree.h"
 #include "pub_core_options.h"
 #include "pub_core_oset.h"
@@ -43,8 +44,9 @@
 #include "pub_core_trampoline.h"
 #include "pub_core_transtab.h"
 #include "pub_core_tooliface.h"    // VG_(needs).malloc_replacement
-#include "pub_tool_machine.h"      // VG_(fnptr_to_fnentry)
+#include "pub_core_machine.h"      // VG_(fnptr_to_fnentry)
 #include "pub_core_aspacemgr.h"    // VG_(am_find_nsegment)
+#include "pub_core_xarray.h"
 #include "pub_core_clientstate.h"  // VG_(client___libc_freeres_wrapper)
 #include "pub_core_demangle.h"     // VG_(maybe_Z_demangle)
 
@@ -204,15 +206,29 @@
 
 /* A specification of a redirection we want to do.  Note that because
    both the "from" soname and function name may contain wildcards, the
-   spec can match an arbitrary number of times. */
+   spec can match an arbitrary number of times. 
+
+   16 Nov 2007: Comments re .mandatory field: The initial motivation
+   for this is making Memcheck work sanely on glibc-2.6.X ppc32-linux.
+   We really need to intercept 'strlen' in ld.so right from startup.
+   If ld.so does not have a visible 'strlen' symbol, Memcheck
+   generates an impossible number of errors resulting from highly
+   tuned strlen implementation in ld.so, and is completely unusable
+   -- the resulting undefinedness eventually seeps everywhere. */
 typedef
    struct _Spec {
       struct _Spec* next;  /* linked list */
+      /* FIXED PARTS -- set when created and not changed */
       HChar* from_sopatt;  /* from soname pattern  */
       HChar* from_fnpatt;  /* from fnname pattern  */
       Addr   to_addr;      /* where redirecting to */
       Bool   isWrap;       /* wrap or replacement? */
-      Bool   mark; /* transient temporary used during matching */
+      const HChar** mandatory; /* non-NULL ==> abort V and print the
+                                  strings if from_sopatt is loaded but
+                                  from_fnpatt cannot be found */
+      /* VARIABLE PARTS -- used transiently whilst processing redirections */
+      Bool   mark; /* set if spec requires further processing */
+      Bool   done; /* set if spec was successfully matched */
    }
    Spec;
 
@@ -224,9 +240,9 @@ typedef
 typedef
    struct _TopSpec {
       struct _TopSpec* next; /* linked list */
-      SegInfo* seginfo;      /* symbols etc */
-      Spec*    specs;        /* specs pulled out of seginfo */
-      Bool     mark; /* transient temporary used during deletion */
+      DebugInfo* seginfo;    /* symbols etc */
+      Spec*      specs;      /* specs pulled out of seginfo */
+      Bool       mark; /* transient temporary used during deletion */
    }
    TopSpec;
 
@@ -265,10 +281,11 @@ static OSet* activeSet = NULL;
 
 static void maybe_add_active ( Active /*by value; callee copies*/ );
 
-static void*  symtab_alloc(SizeT);
-static void   symtab_free(void*);
-static HChar* symtab_strdup(HChar*);
+static void*  dinfo_zalloc(HChar* ec, SizeT);
+static void   dinfo_free(void*);
+static HChar* dinfo_strdup(HChar* ec, HChar*);
 static Bool   is_plausible_guest_addr(Addr);
+static Bool   is_aix5_glink_idiom(Addr);
 
 static void   show_redir_state ( HChar* who );
 static void   show_active ( HChar* left, Active* act );
@@ -286,61 +303,77 @@ void generate_and_add_actives (
         /* spec list and the owning TopSpec */
         Spec*    specs, 
         TopSpec* parent_spec,
-	/* seginfo and the owning TopSpec */
-        SegInfo* si,
+	/* debuginfo and the owning TopSpec */
+        DebugInfo* di,
         TopSpec* parent_sym 
      );
 
-/* Notify m_redir of the arrival of a new SegInfo.  This is fairly
+/* Notify m_redir of the arrival of a new DebugInfo.  This is fairly
    complex, but the net effect is to (1) add a new entry to the
    topspecs list, and (2) figure out what new binding are now active,
    and, as a result, add them to the actives mapping. */
 
 #define N_DEMANGLED 256
 
-void VG_(redir_notify_new_SegInfo)( SegInfo* newsi )
+void VG_(redir_notify_new_DebugInfo)( DebugInfo* newsi )
 {
-   Bool     ok, isWrap;
-   Int      i, nsyms;
-   Spec*    specList;
-   Spec*    spec;
-   TopSpec* ts;
-   TopSpec* newts;
-   HChar*   sym_name;
-   Addr     sym_addr;
-   HChar    demangled_sopatt[N_DEMANGLED];
-   HChar    demangled_fnpatt[N_DEMANGLED];
-
+   Bool         ok, isWrap;
+   Int          i, nsyms;
+   Spec*        specList;
+   Spec*        spec;
+   TopSpec*     ts;
+   TopSpec*     newts;
+   HChar*       sym_name;
+   Addr         sym_addr, sym_toc;
+   HChar        demangled_sopatt[N_DEMANGLED];
+   HChar        demangled_fnpatt[N_DEMANGLED];
+   Bool         check_ppcTOCs = False;
+   Bool         isText;
    const UChar* newsi_soname;
 
+#  if defined(VG_PLAT_USES_PPCTOC)
+   check_ppcTOCs = True;
+#  endif
+
    vg_assert(newsi);
-   newsi_soname = VG_(seginfo_soname)(newsi);
+   newsi_soname = VG_(DebugInfo_get_soname)(newsi);
    vg_assert(newsi_soname != NULL);
 
    /* stay sane: we don't already have this. */
    for (ts = topSpecs; ts; ts = ts->next)
       vg_assert(ts->seginfo != newsi);
 
-   /* scan this SegInfo's symbol table, pulling out and demangling
+   /* scan this DebugInfo's symbol table, pulling out and demangling
       any specs found */
 
    specList = NULL; /* the spec list we're building up */
 
-   nsyms = VG_(seginfo_syms_howmany)( newsi );
+   nsyms = VG_(DebugInfo_syms_howmany)( newsi );
    for (i = 0; i < nsyms; i++) {
-      VG_(seginfo_syms_getidx)( newsi, i, &sym_addr, NULL, &sym_name );
+      VG_(DebugInfo_syms_getidx)( newsi, i, &sym_addr, &sym_toc, 
+                                            NULL, &sym_name, &isText );
       ok = VG_(maybe_Z_demangle)( sym_name, demangled_sopatt, N_DEMANGLED,
-				  demangled_fnpatt, N_DEMANGLED, &isWrap );
+                                  demangled_fnpatt, N_DEMANGLED, &isWrap );
+      /* ignore data symbols */
+      if (!isText)
+         continue;
       if (!ok) {
          /* It's not a full-scale redirect, but perhaps it is a load-notify
             fn?  Let the load-notify department see it. */
          handle_maybe_load_notifier( newsi_soname, sym_name, sym_addr );
          continue; 
       }
-      spec = symtab_alloc(sizeof(Spec));
+      if (check_ppcTOCs && sym_toc == 0) {
+         /* This platform uses toc pointers, but none could be found
+            for this symbol, so we can't safely redirect/wrap to it.
+            Just skip it; we'll make a second pass over the symbols in
+            the following loop, and complain at that point. */
+         continue;
+      }
+      spec = dinfo_zalloc("redir.rnnD.1", sizeof(Spec));
       vg_assert(spec);
-      spec->from_sopatt = symtab_strdup(demangled_sopatt);
-      spec->from_fnpatt = symtab_strdup(demangled_fnpatt);
+      spec->from_sopatt = dinfo_strdup("redir.rnnD.2", demangled_sopatt);
+      spec->from_fnpatt = dinfo_strdup("redir.rnnD.3", demangled_fnpatt);
       vg_assert(spec->from_sopatt);
       vg_assert(spec->from_fnpatt);
       spec->to_addr = sym_addr;
@@ -349,12 +382,44 @@ void VG_(redir_notify_new_SegInfo)( SegInfo* newsi )
       vg_assert(is_plausible_guest_addr(sym_addr));
       spec->next = specList;
       spec->mark = False; /* not significant */
+      spec->done = False; /* not significant */
       specList = spec;
    }
 
-   /* Ok.  Now specList holds the list of specs from the SegInfo. 
+   if (check_ppcTOCs) {
+      for (i = 0; i < nsyms; i++) {
+         VG_(DebugInfo_syms_getidx)( newsi, i, &sym_addr, &sym_toc, 
+                                               NULL, &sym_name, &isText );
+         ok = isText
+              && VG_(maybe_Z_demangle)( 
+                    sym_name, demangled_sopatt, N_DEMANGLED,
+                    demangled_fnpatt, N_DEMANGLED, &isWrap );
+         if (!ok)
+            /* not a redirect.  Ignore. */
+            continue;
+         if (sym_toc != 0)
+            /* has a valid toc pointer.  Ignore. */
+            continue;
+
+         for (spec = specList; spec; spec = spec->next) 
+            if (0 == VG_(strcmp)(spec->from_sopatt, demangled_sopatt)
+                && 0 == VG_(strcmp)(spec->from_fnpatt, demangled_fnpatt))
+               break;
+         if (spec)
+            /* a redirect to some other copy of that symbol, which
+               does have a TOC value, already exists */
+            continue;
+
+         /* Complain */
+         VG_(message)(Vg_DebugMsg,
+                      "WARNING: no TOC ptr for redir/wrap to %s %s\n",
+                      demangled_sopatt, demangled_fnpatt);
+      }
+   }
+
+   /* Ok.  Now specList holds the list of specs from the DebugInfo. 
       Build a new TopSpec, but don't add it to topSpecs yet. */
-   newts = symtab_alloc(sizeof(TopSpec));
+   newts = dinfo_zalloc("redir.rnnD.4", sizeof(TopSpec));
    vg_assert(newts);
    newts->next    = NULL; /* not significant */
    newts->seginfo = newsi;
@@ -384,7 +449,7 @@ void VG_(redir_notify_new_SegInfo)( SegInfo* newsi )
          generate_and_add_actives( specList,    newts,
                                    ts->seginfo, ts );
    }
-	
+
    /* Case (2) */
    for (ts = topSpecs; ts; ts = ts->next) {
       generate_and_add_actives( ts->specs, ts, 
@@ -400,7 +465,7 @@ void VG_(redir_notify_new_SegInfo)( SegInfo* newsi )
    topSpecs = newts;
 
    if (VG_(clo_trace_redir))
-      show_redir_state("after VG_(redir_notify_new_SegInfo)");
+      show_redir_state("after VG_(redir_notify_new_DebugInfo)");
 }
 
 #undef N_DEMANGLED
@@ -417,23 +482,25 @@ void generate_and_add_actives (
         Spec*    specs, 
         TopSpec* parent_spec,
 	/* seginfo and the owning TopSpec */
-        SegInfo* si,
+        DebugInfo* di,
         TopSpec* parent_sym 
      )
 {
    Spec*  sp;
-   Bool   anyMark;
+   Bool   anyMark, isText;
    Active act;
    Int    nsyms, i;
    Addr   sym_addr;
    HChar* sym_name;
 
-   /* First figure out which of the specs match the seginfo's
-      soname. */
+   /* First figure out which of the specs match the seginfo's soname.
+      Also clear the 'done' bits, so that after the main loop below
+      tell which of the Specs really did get done. */
    anyMark = False;
    for (sp = specs; sp; sp = sp->next) {
+      sp->done = False;
       sp->mark = VG_(string_match)( sp->from_sopatt, 
-                                    VG_(seginfo_soname)(si) );
+                                    VG_(DebugInfo_get_soname)(di) );
       anyMark = anyMark || sp->mark;
    }
 
@@ -443,9 +510,25 @@ void generate_and_add_actives (
 
    /* Iterate outermost over the symbols in the seginfo, in the hope
       of trashing the caches less. */
-   nsyms = VG_(seginfo_syms_howmany)( si );
+   nsyms = VG_(DebugInfo_syms_howmany)( di );
    for (i = 0; i < nsyms; i++) {
-      VG_(seginfo_syms_getidx)( si, i, &sym_addr, NULL, &sym_name );
+      VG_(DebugInfo_syms_getidx)( di, i, &sym_addr, NULL, NULL,
+                                         &sym_name, &isText );
+
+      /* ignore data symbols */
+      if (!isText)
+         continue;
+
+      /* On AIX, we cannot redirect calls to a so-called glink
+         function for reasons which are not obvious - something to do
+         with saving r2 across the call.  Not a problem, as we don't
+         want to anyway; presumably it is the target of the glink we
+         need to redirect.  Hence just spot them and ignore them.
+         They are always of a very specific (more or less
+         ABI-mandated) form. */
+      if (is_aix5_glink_idiom(sym_addr))
+         continue;
+
       for (sp = specs; sp; sp = sp->next) {
          if (!sp->mark)
             continue; /* soname doesn't match */
@@ -456,9 +539,60 @@ void generate_and_add_actives (
             act.parent_spec = parent_spec;
             act.parent_sym  = parent_sym;
             act.isWrap      = sp->isWrap;
+            sp->done = True;
             maybe_add_active( act );
          }
-      }
+      } /* for (sp = specs; sp; sp = sp->next) */
+   } /* for (i = 0; i < nsyms; i++)  */
+
+   /* Now, finally, look for Specs which were marked to be done, but
+      didn't get matched.  If any such are mandatory we must abort the
+      system at this point. */
+   for (sp = specs; sp; sp = sp->next) {
+      if (!sp->mark)
+         continue;
+      if (sp->mark && (!sp->done) && sp->mandatory)
+         break;
+   }
+   if (sp) {
+      const HChar** strp;
+      HChar* v = "valgrind:  ";
+      vg_assert(sp->mark);
+      vg_assert(!sp->done);
+      vg_assert(sp->mandatory);
+      VG_(printf)("\n");
+      VG_(printf)(
+      "%sFatal error at startup: a function redirection\n", v);
+      VG_(printf)(
+      "%swhich is mandatory for this platform-tool combination\n", v);
+      VG_(printf)(
+      "%scannot be set up.  Details of the redirection are:\n", v);
+      VG_(printf)(
+      "%s\n", v);
+      VG_(printf)(
+      "%sA must-be-redirected function\n", v);
+      VG_(printf)(
+      "%swhose name matches the pattern:      %s\n", v, sp->from_fnpatt);
+      VG_(printf)(
+      "%sin an object with soname matching:   %s\n", v, sp->from_sopatt);
+      VG_(printf)(
+      "%swas not found whilst processing\n", v);
+      VG_(printf)(
+      "%ssymbols from the object with soname: %s\n",
+      v, VG_(DebugInfo_get_soname)(di));
+      VG_(printf)(
+      "%s\n", v);
+
+      for (strp = sp->mandatory; *strp; strp++)
+         VG_(printf)(
+         "%s%s\n", v, *strp);
+
+      VG_(printf)(
+      "%s\n", v);
+      VG_(printf)(
+      "%sCannot continue -- exiting now.  Sorry.\n", v);
+      VG_(printf)("\n");
+      VG_(exit)(1);
    }
 }
 
@@ -490,7 +624,7 @@ static void maybe_add_active ( Active act )
       goto bad;
    }
 
-   old = VG_(OSet_Lookup)( activeSet, &act.from_addr );
+   old = VG_(OSetGen_Lookup)( activeSet, &act.from_addr );
    if (old) {
       /* Dodgy.  Conflicting binding. */
       vg_assert(old->from_addr == act.from_addr);
@@ -505,36 +639,36 @@ static void maybe_add_active ( Active act )
          /* XXXXXXXXXXX COMPLAIN if new and old parents differ */
       }
    } else {
-      Active* a = VG_(OSet_AllocNode)(activeSet, sizeof(Active));
+      Active* a = VG_(OSetGen_AllocNode)(activeSet, sizeof(Active));
       vg_assert(a);
       *a = act;
-      VG_(OSet_Insert)(activeSet, a);
+      VG_(OSetGen_Insert)(activeSet, a);
       /* Now that a new from->to redirection is in force, we need to
          get rid of any translations intersecting 'from' in order that
          they get redirected to 'to'.  So discard them.  Just for
          paranoia (but, I believe, unnecessarily), discard 'to' as
          well. */
       VG_(discard_translations)( (Addr64)act.from_addr, 1,
-                                 "redir_new_SegInfo(from_addr)");
+                                 "redir_new_DebugInfo(from_addr)");
       VG_(discard_translations)( (Addr64)act.to_addr, 1,
-                                 "redir_new_SegInfo(to_addr)");
+                                 "redir_new_DebugInfo(to_addr)");
    }
    return;
 
   bad:
    vg_assert(what);
    if (VG_(clo_verbosity) > 1) {
-      VG_(message)(Vg_UserMsg, "WARNING: %s", what);
+      VG_(message)(Vg_UserMsg, "WARNING: %s\n", what);
       show_active(             "    new: ", &act);
    }
 }
 
 
-/* Notify m_redir of the deletion of a SegInfo.  This is relatively
+/* Notify m_redir of the deletion of a DebugInfo.  This is relatively
    simple -- just get rid of all actives derived from it, and free up
    the associated list elements. */
 
-void VG_(redir_notify_delete_SegInfo)( SegInfo* delsi )
+void VG_(redir_notify_delete_DebugInfo)( DebugInfo* delsi )
 {
    TopSpec* ts;
    TopSpec* tsPrev;
@@ -543,7 +677,7 @@ void VG_(redir_notify_delete_SegInfo)( SegInfo* delsi )
    OSet*    tmpSet;
    Active*  act;
    Bool     delMe;
-   Addr*    addrP;
+   Addr     addr;
 
    vg_assert(delsi);
 
@@ -558,18 +692,17 @@ void VG_(redir_notify_delete_SegInfo)( SegInfo* delsi )
      ts = ts->next;
    }
 
-   vg_assert(ts); /* else we don't have the deleted SegInfo */
+   vg_assert(ts); /* else we don't have the deleted DebugInfo */
    vg_assert(ts->seginfo == delsi);
 
    /* Traverse the actives, copying the addresses of those we intend
       to delete into tmpSet. */
-   tmpSet = VG_(OSet_Create)( 0/*keyOff*/, NULL/*fastCmp*/,
-                              symtab_alloc, symtab_free);
+   tmpSet = VG_(OSetWord_Create)(dinfo_zalloc, "redir.rndD.1", dinfo_free);
 
    ts->mark = True;
 
-   VG_(OSet_ResetIter)( activeSet );
-   while ( (act = VG_(OSet_Next)(activeSet)) ) {
+   VG_(OSetGen_ResetIter)( activeSet );
+   while ( (act = VG_(OSetGen_Next)(activeSet)) ) {
       delMe = act->parent_spec != NULL
               && act->parent_sym != NULL
               && act->parent_spec->seginfo != NULL
@@ -590,36 +723,33 @@ void VG_(redir_notify_delete_SegInfo)( SegInfo* delsi )
       }
 
       if (delMe) {
-         addrP = VG_(OSet_AllocNode)( tmpSet, sizeof(Addr) );
-         *addrP = act->from_addr;
-         VG_(OSet_Insert)( tmpSet, addrP );
+         VG_(OSetWord_Insert)( tmpSet, act->from_addr );
          /* While we have our hands on both the 'from' and 'to'
             of this Active, do paranoid stuff with tt/tc. */
          VG_(discard_translations)( (Addr64)act->from_addr, 1,
-                                    "redir_del_SegInfo(from_addr)");
+                                    "redir_del_DebugInfo(from_addr)");
          VG_(discard_translations)( (Addr64)act->to_addr, 1,
-                                    "redir_del_SegInfo(to_addr)");
+                                    "redir_del_DebugInfo(to_addr)");
       }
    }
 
-   /* Now traverse tmpSet, deleting corresponding elements in
-      activeSet. */
-   VG_(OSet_ResetIter)( tmpSet );
-   while ( (addrP = VG_(OSet_Next)(tmpSet)) ) {
-      act = VG_(OSet_Remove)( activeSet, addrP );
+   /* Now traverse tmpSet, deleting corresponding elements in activeSet. */
+   VG_(OSetWord_ResetIter)( tmpSet );
+   while ( VG_(OSetWord_Next)(tmpSet, &addr) ) {
+      act = VG_(OSetGen_Remove)( activeSet, &addr );
       vg_assert(act);
-      VG_(OSet_FreeNode)( activeSet, act );
+      VG_(OSetGen_FreeNode)( activeSet, act );
    }
 
-   VG_(OSet_Destroy)( tmpSet, NULL );
+   VG_(OSetWord_Destroy)( tmpSet );
 
    /* The Actives set is now cleaned up.  Free up this TopSpec and
       everything hanging off it. */
    for (sp = ts->specs; sp; sp = sp_next) {
-      if (sp->from_sopatt) symtab_free(sp->from_sopatt);
-      if (sp->from_fnpatt) symtab_free(sp->from_fnpatt);
+      if (sp->from_sopatt) dinfo_free(sp->from_sopatt);
+      if (sp->from_fnpatt) dinfo_free(sp->from_fnpatt);
       sp_next = sp->next;
-      symtab_free(sp);
+      dinfo_free(sp);
    }
 
    if (tsPrev == NULL) {
@@ -628,10 +758,10 @@ void VG_(redir_notify_delete_SegInfo)( SegInfo* delsi )
    } else {
       tsPrev->next = ts->next;
    }
-   symtab_free(ts);
+   dinfo_free(ts);
 
    if (VG_(clo_trace_redir))
-      show_redir_state("after VG_(redir_notify_delete_SegInfo)");
+      show_redir_state("after VG_(redir_notify_delete_DebugInfo)");
 }
 
 
@@ -644,7 +774,7 @@ void VG_(redir_notify_delete_SegInfo)( SegInfo* delsi )
    just before translating a basic block. */
 Addr VG_(redir_do_lookup) ( Addr orig, Bool* isWrap )
 {
-   Active* r = VG_(OSet_Lookup)(activeSet, &orig);
+   Active* r = VG_(OSetGen_Lookup)(activeSet, &orig);
    if (r == NULL)
       return orig;
 
@@ -681,51 +811,66 @@ static void add_hardwired_active ( Addr from, Addr to )
    entry that holds these initial specs. */
 
 __attribute__((unused)) /* not used on all platforms */
-static void add_hardwired_spec ( HChar* sopatt, HChar* fnpatt, Addr to_addr )
+static void add_hardwired_spec ( HChar* sopatt, HChar* fnpatt, 
+                                 Addr   to_addr,
+                                 const HChar** mandatory )
 {
-   Spec* spec = symtab_alloc(sizeof(Spec));
+   Spec* spec = dinfo_zalloc("redir.ahs.1", sizeof(Spec));
    vg_assert(spec);
 
    if (topSpecs == NULL) {
-      topSpecs = symtab_alloc(sizeof(TopSpec));
+      topSpecs = dinfo_zalloc("redir.ahs.2", sizeof(TopSpec));
       vg_assert(topSpecs);
-      topSpecs->next    = NULL;
-      topSpecs->seginfo = NULL;
-      topSpecs->specs   = NULL;
-      topSpecs->mark    = False;
+      /* symtab_zalloc sets all fields to zero */
    }
 
    vg_assert(topSpecs != NULL);
    vg_assert(topSpecs->next == NULL);
    vg_assert(topSpecs->seginfo == NULL);
-
+   /* FIXED PARTS */
    spec->from_sopatt = sopatt;
    spec->from_fnpatt = fnpatt;
    spec->to_addr     = to_addr;
    spec->isWrap      = False;
+   spec->mandatory   = mandatory;
+   /* VARIABLE PARTS */
    spec->mark        = False; /* not significant */
+   spec->done        = False; /* not significant */
 
    spec->next = topSpecs->specs;
    topSpecs->specs = spec;
 }
 
 
+__attribute__((unused)) /* not used on all platforms */
+static const HChar* complain_about_stripped_glibc_ldso[]
+= { "Possible fixes: (1, short term): install glibc's debuginfo",
+    "package on this machine.  (2, longer term): ask the packagers",
+    "for your Linux distribution to please in future ship a non-",
+    "stripped ld.so (or whatever the dynamic linker .so is called)",
+    "that exports the above-named function using the standard",
+    "calling conventions for this platform.",
+    NULL
+  };
+
+
 /* Initialise the redir system, and create the initial Spec list and
    for amd64-linux a couple of permanent active mappings.  The initial
    Specs are not converted into Actives yet, on the (checked)
-   assumption that no SegInfos have so far been created, and so when
+   assumption that no DebugInfos have so far been created, and so when
    they are created, that will happen. */
 
 void VG_(redir_initialise) ( void )
 {
-   // Assert that there are no SegInfos so far
-   vg_assert( VG_(next_seginfo)(NULL) == NULL );
+   // Assert that there are no DebugInfos so far
+   vg_assert( VG_(next_DebugInfo)(NULL) == NULL );
 
    // Initialise active mapping.
-   activeSet = VG_(OSet_Create)(offsetof(Active, from_addr),
-                                NULL,     // Use fast comparison
-                                symtab_alloc,
-                                symtab_free);
+   activeSet = VG_(OSetGen_Create)(offsetof(Active, from_addr),
+                                   NULL,     // Use fast comparison
+                                   dinfo_zalloc,
+                                   "redir.ri.1", 
+                                   dinfo_free);
 
    // The rest of this function just adds initial Specs.   
 
@@ -735,7 +880,8 @@ void VG_(redir_initialise) ( void )
    if (0==VG_(strcmp)("Memcheck", VG_(details).name)) {
       add_hardwired_spec(
          "ld-linux.so.2", "index",
-          (Addr)&VG_(x86_linux_REDIR_FOR_index)
+         (Addr)&VG_(x86_linux_REDIR_FOR_index),
+         NULL
       );
    }
 
@@ -750,21 +896,47 @@ void VG_(redir_initialise) ( void )
       (Addr)&VG_(amd64_linux_REDIR_FOR_vtime) 
    );
 
+   /* If we're using memcheck, use these intercepts right from
+      the start, otherwise ld.so makes a lot of noise. */
+   if (0==VG_(strcmp)("Memcheck", VG_(details).name)) {
+
+      add_hardwired_spec(
+         "ld-linux-x86-64.so.2", "strlen",
+         (Addr)&VG_(amd64_linux_REDIR_FOR_strlen),
+#        if defined(GLIBC_2_2) || defined(GLIBC_2_3) || defined(GLIBC_2_4) \
+            || defined(GLIBC_2_5) || defined(GLIBC_2_6) || defined(GLIBC_2_7) \
+            || defined(GLIBC_2_8) || defined(GLIBC_2_9)
+         NULL
+#        else
+         /* for glibc-2.10 and later, this is mandatory - can't sanely
+            continue without it */
+         complain_about_stripped_glibc_ldso
+#        endif
+      );   
+   }
+
 #  elif defined(VGP_ppc32_linux)
    /* If we're using memcheck, use these intercepts right from
       the start, otherwise ld.so makes a lot of noise. */
    if (0==VG_(strcmp)("Memcheck", VG_(details).name)) {
+
+      /* this is mandatory - can't sanely continue without it */
       add_hardwired_spec(
          "ld.so.1", "strlen",
-          (Addr)&VG_(ppc32_linux_REDIR_FOR_strlen)
+         (Addr)&VG_(ppc32_linux_REDIR_FOR_strlen),
+         complain_about_stripped_glibc_ldso
       );   
       add_hardwired_spec(
          "ld.so.1", "strcmp",
-         (Addr)&VG_(ppc32_linux_REDIR_FOR_strcmp)
+         (Addr)&VG_(ppc32_linux_REDIR_FOR_strcmp),
+         NULL /* not mandatory - so why bother at all? */
+         /* glibc-2.6.1 (openSUSE 10.3, ppc32) seems fine without it */
       );
       add_hardwired_spec(
          "ld.so.1", "index",
-         (Addr)&VG_(ppc32_linux_REDIR_FOR_strchr)
+         (Addr)&VG_(ppc32_linux_REDIR_FOR_strchr),
+         NULL /* not mandatory - so why bother at all? */
+         /* glibc-2.6.1 (openSUSE 10.3, ppc32) seems fine without it */
       );
    }
 
@@ -773,16 +945,46 @@ void VG_(redir_initialise) ( void )
       the start, otherwise ld.so makes a lot of noise. */
    if (0==VG_(strcmp)("Memcheck", VG_(details).name)) {
 
+      /* this is mandatory - can't sanely continue without it */
       add_hardwired_spec(
          "ld64.so.1", "strlen",
-         (Addr)VG_(fnptr_to_fnentry)( &VG_(ppc64_linux_REDIR_FOR_strlen) )
-      );   
+         (Addr)VG_(fnptr_to_fnentry)( &VG_(ppc64_linux_REDIR_FOR_strlen) ),
+         complain_about_stripped_glibc_ldso
+      );
 
       add_hardwired_spec(
          "ld64.so.1", "index",
-         (Addr)VG_(fnptr_to_fnentry)( &VG_(ppc64_linux_REDIR_FOR_strchr) )
-      );   
+         (Addr)VG_(fnptr_to_fnentry)( &VG_(ppc64_linux_REDIR_FOR_strchr) ),
+         NULL /* not mandatory - so why bother at all? */
+         /* glibc-2.5 (FC6, ppc64) seems fine without it */
+      );
+   }
 
+#  elif defined(VGP_ppc32_aix5)
+   /* nothing so far */
+
+#  elif defined(VGP_ppc64_aix5)
+   /* nothing so far */
+
+#  elif defined(VGO_darwin)
+   /* If we're using memcheck, use these intercepts right from
+      the start, otherwise dyld makes a lot of noise. */
+   if (0==VG_(strcmp)("Memcheck", VG_(details).name)) {
+      add_hardwired_spec("dyld", "strcmp",
+                         (Addr)&VG_(darwin_REDIR_FOR_strcmp), NULL);
+      add_hardwired_spec("dyld", "strlen",
+                         (Addr)&VG_(darwin_REDIR_FOR_strlen), NULL);
+      add_hardwired_spec("dyld", "strcat",
+                         (Addr)&VG_(darwin_REDIR_FOR_strcat), NULL);
+      add_hardwired_spec("dyld", "strcpy",
+                         (Addr)&VG_(darwin_REDIR_FOR_strcpy), NULL);
+      add_hardwired_spec("dyld", "strlcat",
+                         (Addr)&VG_(darwin_REDIR_FOR_strlcat), NULL);
+#     if defined(VGP_amd64_darwin)
+      // DDD: #warning fixme rdar://6166275
+      add_hardwired_spec("dyld", "arc4random",
+                         (Addr)&VG_(darwin_REDIR_FOR_arc4random), NULL);
+#     endif
    }
 
 #  else
@@ -798,31 +1000,71 @@ void VG_(redir_initialise) ( void )
 /*--- MISC HELPERS                                         ---*/
 /*------------------------------------------------------------*/
 
-static void* symtab_alloc(SizeT n)
-{
-   return VG_(arena_malloc)(VG_AR_SYMTAB, n);
+static void* dinfo_zalloc(HChar* ec, SizeT n) {
+   void* p;
+   vg_assert(n > 0);
+   p = VG_(arena_malloc)(VG_AR_DINFO, ec, n);
+   tl_assert(p);
+   VG_(memset)(p, 0, n);
+   return p;
 }
 
-static void symtab_free(void* p)
-{
-   return VG_(arena_free)(VG_AR_SYMTAB, p);
+static void dinfo_free(void* p) {
+   tl_assert(p);
+   return VG_(arena_free)(VG_AR_DINFO, p);
 }
 
-static HChar* symtab_strdup(HChar* str)
+static HChar* dinfo_strdup(HChar* ec, HChar* str)
 {
-   return VG_(arena_strdup)(VG_AR_SYMTAB, str);
+   return VG_(arena_strdup)(VG_AR_DINFO, ec, str);
 }
 
 /* Really this should be merged with translations_allowable_from_seg
    in m_translate. */
 static Bool is_plausible_guest_addr(Addr a)
 {
-   NSegment* seg = VG_(am_find_nsegment)(a);
+   NSegment const* seg = VG_(am_find_nsegment)(a);
    return seg != NULL
           && (seg->kind == SkAnonC || seg->kind == SkFileC)
           && (seg->hasX || seg->hasR); /* crude x86-specific hack */
 }
 
+/* A function which spots AIX 'glink' functions.  A 'glink' function
+   is a stub function which has something to do with AIX-style dynamic
+   linking, and jumps to the real target (with which it typically
+   shares the same name).  See also comment where this function is
+   used (above). */
+static Bool is_aix5_glink_idiom ( Addr sym_addr )
+{
+#  if defined(VGP_ppc32_aix5)
+   UInt* w = (UInt*)sym_addr;
+   if (VG_IS_4_ALIGNED(w)
+       && is_plausible_guest_addr((Addr)(w+0))
+       && is_plausible_guest_addr((Addr)(w+6))
+       && (w[0] & 0xFFFF0000) == 0x81820000 /* lwz r12,func@toc(r2) */
+       && w[1] == 0x90410014                /* stw r2,20(r1) */
+       && w[2] == 0x800c0000                /* lwz r0,0(r12) */
+       && w[3] == 0x804c0004                /* lwz r2,4(r12) */
+       && w[4] == 0x7c0903a6                /* mtctr r0 */
+       && w[5] == 0x4e800420                /* bctr */
+       && w[6] == 0x00000000                /* illegal */)
+      return True;
+#  elif defined(VGP_ppc64_aix5)
+   UInt* w = (UInt*)sym_addr;
+   if (VG_IS_4_ALIGNED(w)
+       && is_plausible_guest_addr((Addr)(w+0))
+       && is_plausible_guest_addr((Addr)(w+6))
+       && (w[0] & 0xFFFF0000) == 0xE9820000 /* ld  r12,func@toc(r2) */
+       && w[1] == 0xF8410028                /* std r2,40(r1) */
+       && w[2] == 0xE80C0000                /* ld  r0,0(r12) */
+       && w[3] == 0xE84C0008                /* ld  r2,8(r12) */
+       && w[4] == 0x7c0903a6                /* mtctr r0 */
+       && w[5] == 0x4e800420                /* bctr */
+       && w[6] == 0x00000000                /* illegal */)
+      return True;
+#  endif
+   return False;
+}
 
 /*------------------------------------------------------------*/
 /*--- NOTIFY-ON-LOAD FUNCTIONS                             ---*/
@@ -865,12 +1107,12 @@ void handle_maybe_load_notifier( const UChar* soname,
 
 static void show_spec ( HChar* left, Spec* spec )
 {
-   VG_(message)(Vg_DebugMsg, 
-                  "%s%18s %30s %s-> 0x%08llx",
-                  left,
-                  spec->from_sopatt, spec->from_fnpatt,
-                  spec->isWrap ? "W" : "R",
-                  (ULong)spec->to_addr );
+   VG_(message)( Vg_DebugMsg, 
+                 "%s%25s %30s %s-> 0x%08llx\n",
+                 left,
+                 spec->from_sopatt, spec->from_fnpatt,
+                 spec->isWrap ? "W" : "R",
+                 (ULong)spec->to_addr );
 }
 
 static void show_active ( HChar* left, Active* act )
@@ -884,7 +1126,7 @@ static void show_active ( HChar* left, Active* act )
    ok = VG_(get_fnname_w_offset)(act->to_addr, name2, 64);
    if (!ok) VG_(strcpy)(name2, "???");
 
-   VG_(message)(Vg_DebugMsg, "%s0x%08llx (%10s) %s-> 0x%08llx %s", 
+   VG_(message)(Vg_DebugMsg, "%s0x%08llx (%20s) %s-> 0x%08llx %s\n", 
                              left, 
                              (ULong)act->from_addr, name1,
                              act->isWrap ? "W" : "R",
@@ -896,23 +1138,24 @@ static void show_redir_state ( HChar* who )
    TopSpec* ts;
    Spec*    sp;
    Active*  act;
-   VG_(message)(Vg_DebugMsg, "<<");
-   VG_(message)(Vg_DebugMsg, "   ------ REDIR STATE %s ------", who);
+   VG_(message)(Vg_DebugMsg, "<<\n");
+   VG_(message)(Vg_DebugMsg, "   ------ REDIR STATE %s ------\n", who);
    for (ts = topSpecs; ts; ts = ts->next) {
       VG_(message)(Vg_DebugMsg, 
-                   "   TOPSPECS of soname %s",
-                   ts->seginfo ? (HChar*)VG_(seginfo_soname)(ts->seginfo)
-                               : "(hardwired)" );
+                   "   TOPSPECS of soname %s\n",
+                   ts->seginfo
+                      ? (HChar*)VG_(DebugInfo_get_soname)(ts->seginfo)
+                      : "(hardwired)" );
       for (sp = ts->specs; sp; sp = sp->next)
          show_spec("     ", sp);
    }
-   VG_(message)(Vg_DebugMsg, "   ------ ACTIVE ------");
-   VG_(OSet_ResetIter)( activeSet );
-   while ( (act = VG_(OSet_Next)(activeSet)) ) {
+   VG_(message)(Vg_DebugMsg, "   ------ ACTIVE ------\n");
+   VG_(OSetGen_ResetIter)( activeSet );
+   while ( (act = VG_(OSetGen_Next)(activeSet)) ) {
       show_active("    ", act);
    }
 
-   VG_(message)(Vg_DebugMsg, ">>");
+   VG_(message)(Vg_DebugMsg, ">>\n");
 }
 
 /*--------------------------------------------------------------------*/
