@@ -7,7 +7,7 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2000-2007 Julian Seward 
+   Copyright (C) 2000-2009 Julian Seward 
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -29,15 +29,18 @@
 */
 
 #include "pub_core_basics.h"
+#include "pub_core_vki.h"
 #include "pub_core_threadstate.h"
-#include "pub_core_debuginfo.h"
+#include "pub_core_debuginfo.h"     // XXX: circular dependency
 #include "pub_core_aspacemgr.h"     // For VG_(is_addressable)()
 #include "pub_core_libcbase.h"
 #include "pub_core_libcassert.h"
 #include "pub_core_libcprint.h"
 #include "pub_core_machine.h"
 #include "pub_core_options.h"
+#include "pub_core_stacks.h"        // VG_(stack_limits)
 #include "pub_core_stacktrace.h"
+#include "pub_core_xarray.h"
 #include "pub_core_clientstate.h"   // VG_(client__dl_sysinfo_int80)
 #include "pub_core_trampoline.h"
 
@@ -45,7 +48,7 @@
 /*--- Exported functions.                                  ---*/
 /*------------------------------------------------------------*/
 
-/* Take a snapshot of the client's stack, putting the up to 'n_ips'
+/* Take a snapshot of the client's stack, putting up to 'max_n_ips'
    IPs into 'ips'.  In order to be thread-safe, we pass in the
    thread's IP SP, FP if that's meaningful, and LR if that's
    meaningful.  Returns number of IPs put in 'ips'.
@@ -54,14 +57,23 @@
    first parameter, else send zero.  This helps generate better stack
    traces on ppc64-linux and has no effect on other platforms.
 */
-UInt VG_(get_StackTrace2) ( ThreadId tid_if_known,
-                            Addr* ips, UInt n_ips, 
-                            Addr ip, Addr sp, Addr fp, Addr lr,
-                            Addr fp_min, Addr fp_max_orig )
+UInt VG_(get_StackTrace_wrk) ( ThreadId tid_if_known,
+                               /*OUT*/Addr* ips, UInt max_n_ips,
+                               /*OUT*/Addr* sps, /*OUT*/Addr* fps,
+                               Addr ip, Addr sp, Addr fp, Addr lr,
+                               Addr fp_min, Addr fp_max_orig )
 {
-#if defined(VGP_ppc32_linux) || defined(VGP_ppc64_linux)
-   Bool  lr_is_first_RA = False; /* ppc only */
-#endif
+#  if defined(VGP_ppc32_linux) || defined(VGP_ppc64_linux) \
+                               || defined(VGP_ppc32_aix5) \
+                               || defined(VGP_ppc64_aix5)
+   Bool  lr_is_first_RA = False;
+#  endif
+#  if defined(VGP_ppc64_linux) || defined(VGP_ppc64_aix5) \
+                               || defined(VGP_ppc32_aix5)
+   Word redir_stack_size = 0;
+   Word redirs_used      = 0;
+#  endif
+
    Bool  debug = False;
    Int   i;
    Addr  fp_max;
@@ -70,7 +82,7 @@ UInt VG_(get_StackTrace2) ( ThreadId tid_if_known,
    vg_assert(sizeof(Addr) == sizeof(UWord));
    vg_assert(sizeof(Addr) == sizeof(void*));
 
-   /* Snaffle IPs from the client's stack into ips[0 .. n_ips-1],
+   /* Snaffle IPs from the client's stack into ips[0 .. max_n_ips-1],
       stopping when the trail goes cold, which we guess to be
       when FP is not a reasonable stack location. */
 
@@ -78,35 +90,43 @@ UInt VG_(get_StackTrace2) ( ThreadId tid_if_known,
    // current page, at least.  Dunno if it helps.
    // NJN 2002-sep-17: seems to -- stack traces look like 1.0.X again
    fp_max = VG_PGROUNDUP(fp_max_orig);
-   fp_max -= sizeof(Addr);
+   if (fp_max >= sizeof(Addr))
+      fp_max -= sizeof(Addr);
 
    if (debug)
-      VG_(printf)("n_ips=%d fp_min=%p fp_max_orig=%p, fp_max=%p ip=%p fp=%p\n",
-		  n_ips, fp_min, fp_max_orig, fp_max, ip, fp);
+      VG_(printf)("max_n_ips=%d fp_min=0x%lx fp_max_orig=0x%lx, "
+                  "fp_max=0x%lx ip=0x%lx fp=0x%lx\n",
+		  max_n_ips, fp_min, fp_max_orig, fp_max, ip, fp);
 
    /* Assertion broken before main() is reached in pthreaded programs;  the
     * offending stack traces only have one item.  --njn, 2002-aug-16 */
    /* vg_assert(fp_min <= fp_max);*/
-
-   if (fp_min + VG_(clo_max_stackframe) <= fp_max) {
-      /* If the stack is ridiculously big, don't poke around ... but
-         don't bomb out either.  Needed to make John Regehr's
-         user-space threads package work. JRS 20021001 */
+   // On Darwin, this kicks in for pthread-related stack traces, so they're
+   // only 1 entry long which is wrong.
+#if !defined(VGO_darwin)
+   if (fp_min + 512 >= fp_max) {
+      /* If the stack limits look bogus, don't poke around ... but
+         don't bomb out either. */
+      if (sps) sps[0] = sp;
+      if (fps) fps[0] = fp;
       ips[0] = ip;
       return 1;
    } 
+#endif
 
    /* Otherwise unwind the stack in a platform-specific way.  Trying
       to merge the x86, amd64, ppc32 and ppc64 logic into a single
       piece of code is just too confusing and difficult to
       performance-tune.  */
 
-#  if defined(VGP_x86_linux)
+#  if defined(VGP_x86_linux) || defined(VGP_x86_darwin)
 
    /*--------------------- x86 ---------------------*/
 
    /* fp is %ebp.  sp is %esp.  ip is %eip. */
 
+   if (sps) sps[0] = sp;
+   if (fps) fps[0] = fp;
    ips[0] = ip;
    i = 1;
 
@@ -127,7 +147,7 @@ UInt VG_(get_StackTrace2) ( ThreadId tid_if_known,
     */
    while (True) {
 
-      if (i >= n_ips)
+      if (i >= max_n_ips)
          break;
 
       /* Try to derive a new (ip,sp,fp) triple from the current
@@ -140,25 +160,51 @@ UInt VG_(get_StackTrace2) ( ThreadId tid_if_known,
          fails, and is expensive. */
       /* Deal with frames resulting from functions which begin "pushl%
          ebp ; movl %esp, %ebp" which is the ABI-mandated preamble. */
-      if (fp_min <= fp && fp <= fp_max) {
+      if (fp_min <= fp &&
+          fp <= fp_max - 1 * sizeof(UWord)/*see comment below*/)
+      {
          /* fp looks sane, so use it. */
          ip = (((UWord*)fp)[1]);
+         // We stop if we hit a zero (the traditional end-of-stack
+         // marker) or a one -- these correspond to recorded IPs of 0 or -1.
+         // The latter because r8818 (in this file) changes the meaning of
+         // entries [1] and above in a stack trace, by subtracting 1 from
+         // them.  Hence stacks that used to end with a zero value now end in
+         // -1 and so we must detect that too.
+         if (0 == ip || 1 == ip) break;
          sp = fp + sizeof(Addr) /*saved %ebp*/ 
                  + sizeof(Addr) /*ra*/;
          fp = (((UWord*)fp)[0]);
-         ips[i++] = ip;
+         if (sps) sps[i] = sp;
+         if (fps) fps[i] = fp;
+         ips[i++] = ip - 1; /* -1: refer to calling insn, not the RA */
          if (debug)
-            VG_(printf)("     ipsF[%d]=%08p\n", i-1, ips[i-1]);
-         ip = ip - 1;
+            VG_(printf)("     ipsF[%d]=0x%08lx\n", i-1, ips[i-1]);
+         ip = ip - 1; /* as per comment at the head of this loop */
          continue;
       }
 
-      /* That didn't work out, so see if there is any CFI info to hand
+      /* That didn't work out, so see if there is any CF info to hand
          which can be used. */
       if ( VG_(use_CF_info)( &ip, &sp, &fp, fp_min, fp_max ) ) {
+         if (0 == ip || 1 == ip) break;
+         if (sps) sps[i] = sp;
+         if (fps) fps[i] = fp;
+         ips[i++] = ip - 1; /* -1: refer to calling insn, not the RA */
+         if (debug)
+            VG_(printf)("     ipsC[%d]=0x%08lx\n", i-1, ips[i-1]);
+         ip = ip - 1; /* as per comment at the head of this loop */
+         continue;
+      }
+
+      /* And, similarly, try for MSVC FPO unwind info. */
+      if ( VG_(use_FPO_info)( &ip, &sp, &fp, fp_min, fp_max ) ) {
+         if (0 == ip || 1 == ip) break;
+         if (sps) sps[i] = sp;
+         if (fps) fps[i] = fp;
          ips[i++] = ip;
          if (debug)
-            VG_(printf)("     ipsC[%d]=%08p\n", i-1, ips[i-1]);
+            VG_(printf)("     ipsC[%d]=0x%08lx\n", i-1, ips[i-1]);
          ip = ip - 1;
          continue;
       }
@@ -167,13 +213,15 @@ UInt VG_(get_StackTrace2) ( ThreadId tid_if_known,
       break;
    }
 
-#  elif defined(VGP_amd64_linux)
+#  elif defined(VGP_amd64_linux)  ||  defined(VGP_amd64_darwin)
 
    /*--------------------- amd64 ---------------------*/
 
    /* fp is %rbp.  sp is %rsp.  ip is %rip. */
 
    ips[0] = ip;
+   if (sps) sps[0] = sp;
+   if (fps) fps[0] = fp;
    i = 1;
 
    /* Loop unwinding the stack. Note that the IP value we get on
@@ -193,19 +241,21 @@ UInt VG_(get_StackTrace2) ( ThreadId tid_if_known,
     */
    while (True) {
 
-      if (i >= n_ips)
+      if (i >= max_n_ips)
          break;
 
-      /* Try to derive a new (ip,sp,fp) triple from the current
-         set. */
+      /* Try to derive a new (ip,sp,fp) triple from the current set. */
 
       /* First off, see if there is any CFI info to hand which can
          be used. */
       if ( VG_(use_CF_info)( &ip, &sp, &fp, fp_min, fp_max ) ) {
-         ips[i++] = ip;
+         if (0 == ip || 1 == ip) break;
+         if (sps) sps[i] = sp;
+         if (fps) fps[i] = fp;
+         ips[i++] = ip - 1; /* -1: refer to calling insn, not the RA */
          if (debug)
-            VG_(printf)("     ipsC[%d]=%08p\n", i-1, ips[i-1]);
-         ip = ip - 1;
+            VG_(printf)("     ipsC[%d]=%#08lx\n", i-1, ips[i-1]);
+         ip = ip - 1; /* as per comment at the head of this loop */
          continue;
       }
 
@@ -217,54 +267,112 @@ UInt VG_(get_StackTrace2) ( ThreadId tid_if_known,
          the start of the fn, like GDB does, there's no reliable way
          to tell.  Hence the hack of first trying out CFI, and if that
          fails, then use this as a fallback. */
-      if (fp_min <= fp && fp <= fp_max) {
+      /* Note: re "- 1 * sizeof(UWord)", need to take account of the
+         fact that we are prodding at & ((UWord*)fp)[1] and so need to
+         adjust the limit check accordingly.  Omitting this has been
+         observed to cause segfaults on rare occasions. */
+      if (fp_min <= fp && fp <= fp_max - 1 * sizeof(UWord)) {
          /* fp looks sane, so use it. */
          ip = (((UWord*)fp)[1]);
+         if (0 == ip || 1 == ip) break;
          sp = fp + sizeof(Addr) /*saved %rbp*/ 
                  + sizeof(Addr) /*ra*/;
          fp = (((UWord*)fp)[0]);
-         ips[i++] = ip;
+         if (sps) sps[i] = sp;
+         if (fps) fps[i] = fp;
+         ips[i++] = ip - 1; /* -1: refer to calling insn, not the RA */
          if (debug)
-            VG_(printf)("     ipsF[%d]=%08p\n", i-1, ips[i-1]);
-         ip = ip - 1;
+            VG_(printf)("     ipsF[%d]=%#08lx\n", i-1, ips[i-1]);
+         ip = ip - 1; /* as per comment at the head of this loop */
          continue;
       }
 
-      /* No luck there.  We have to give up. */
+      /* Last-ditch hack (evidently GDB does something similar).  We
+         are in the middle of nowhere and we have a nonsense value for
+         the frame pointer.  If the stack pointer is still valid,
+         assume that what it points at is a return address.  Yes,
+         desperate measures.  Could do better here:
+         - check that the supposed return address is in
+           an executable page
+         - check that the supposed return address is just after a call insn
+         - given those two checks, don't just consider *sp as the return 
+           address; instead scan a likely section of stack (eg sp .. sp+256)
+           and use suitable values found there.
+      */
+      if (fp_min <= sp && sp < fp_max) {
+         ip = ((UWord*)sp)[0];
+         if (0 == ip || 1 == ip) break;
+         if (sps) sps[i] = sp;
+         if (fps) fps[i] = fp;
+         ips[i++] = ip == 0 
+                    ? 0 /* sp[0] == 0 ==> stuck at the bottom of a
+                           thread stack */
+                    : ip - 1; /* -1: refer to calling insn, not the RA */
+         if (debug)
+            VG_(printf)("     ipsH[%d]=%#08lx\n", i-1, ips[i-1]);
+         ip = ip - 1; /* as per comment at the head of this loop */
+         sp += 8;
+         continue;
+      }
+
+      /* No luck at all.  We have to give up. */
       break;
    }
 
-#  elif defined(VGP_ppc32_linux) || defined(VGP_ppc64_linux)
+#  elif defined(VGP_ppc32_linux) || defined(VGP_ppc64_linux) \
+        || defined(VGP_ppc32_aix5) || defined(VGP_ppc64_aix5)
 
    /*--------------------- ppc32/64 ---------------------*/
 
    /* fp is %r1.  ip is %cia.  Note, ppc uses r1 as both the stack and
       frame pointers. */
 
-#  if defined(VGP_ppc64_linux)
+#  if defined(VGP_ppc64_linux) || defined(VGP_ppc64_aix5)
+   redir_stack_size = VEX_GUEST_PPC64_REDIR_STACK_SIZE;
+   redirs_used      = 0;
+#  elif defined(VGP_ppc32_aix5)
+   redir_stack_size = VEX_GUEST_PPC32_REDIR_STACK_SIZE;
+   redirs_used      = 0;
+#  endif
+
+#  if defined(VG_PLAT_USES_PPCTOC)
    /* Deal with bogus LR values caused by function
-      interception/wrapping; see comment on similar code a few lines
-      further down. */
-   if (lr == (Addr)&VG_(ppc64_linux_magic_redirect_return_stub)
+      interception/wrapping on ppc-TOC platforms; see comment on
+      similar code a few lines further down. */
+   if (ULong_to_Ptr(lr) == (void*)&VG_(ppctoc_magic_redirect_return_stub)
        && VG_(is_valid_tid)(tid_if_known)) {
-      Long hsp = VG_(threads)[tid_if_known].arch.vex.guest_REDIR_SP;
-      if (hsp >= 1 && hsp < VEX_GUEST_PPC64_REDIR_STACK_SIZE)
+      Word hsp = VG_(threads)[tid_if_known].arch.vex.guest_REDIR_SP;
+      redirs_used++;
+      if (hsp >= 1 && hsp < redir_stack_size)
          lr = VG_(threads)[tid_if_known]
                  .arch.vex.guest_REDIR_STACK[hsp-1];
    }
 #  endif
 
+   /* We have to determine whether or not LR currently holds this fn
+      (call it F)'s return address.  It might not if F has previously
+      called some other function, hence overwriting LR with a pointer
+      to some part of F.  Hence if LR and IP point to the same
+      function then we conclude LR does not hold this function's
+      return address; instead the LR at entry must have been saved in
+      the stack by F's prologue and so we must get it from there
+      instead.  Note all this guff only applies to the innermost
+      frame. */
    lr_is_first_RA = False;
    {
 #     define M_VG_ERRTXT 1000
       UChar buf_lr[M_VG_ERRTXT], buf_ip[M_VG_ERRTXT];
-      if (VG_(get_fnname_nodemangle) (lr, buf_lr, M_VG_ERRTXT))
-         if (VG_(get_fnname_nodemangle) (ip, buf_ip, M_VG_ERRTXT))
+      /* The following conditional looks grossly inefficient and
+         surely could be majorly improved, with not much effort. */
+      if (VG_(get_fnname_raw) (lr, buf_lr, M_VG_ERRTXT))
+         if (VG_(get_fnname_raw) (ip, buf_ip, M_VG_ERRTXT))
             if (VG_(strncmp)(buf_lr, buf_ip, M_VG_ERRTXT))
                lr_is_first_RA = True;
 #     undef M_VG_ERRTXT
    }
 
+   if (sps) sps[0] = fp; /* NB. not sp */
+   if (fps) fps[0] = fp;
    ips[0] = ip;
    i = 1;
 
@@ -275,21 +383,22 @@ UInt VG_(get_StackTrace2) ( ThreadId tid_if_known,
 
       while (True) {
 
-        /* on ppc64-linux (ppc64-elf, really), the lr save slot is 2
-           words back from sp, whereas on ppc32-elf(?) it's only one
-           word back. */
-#        if defined(VGP_ppc64_linux)
+        /* On ppc64-linux (ppc64-elf, really), and on AIX, the lr save
+           slot is 2 words back from sp, whereas on ppc32-elf(?) it's
+           only one word back. */
+#        if defined(VGP_ppc64_linux) \
+            || defined(VGP_ppc32_aix5) || defined(VGP_ppc64_aix5)
          const Int lr_offset = 2;
 #        else
          const Int lr_offset = 1;
 #        endif
 
-         if (i >= n_ips)
+         if (i >= max_n_ips)
             break;
 
          /* Try to derive a new (ip,fp) pair from the current set. */
 
-         if (fp_min <= fp && fp <= fp_max) {
+         if (fp_min <= fp && fp <= fp_max - lr_offset * sizeof(UWord)) {
             /* fp looks sane, so use it. */
 
             if (i == 1 && lr_is_first_RA)
@@ -297,28 +406,39 @@ UInt VG_(get_StackTrace2) ( ThreadId tid_if_known,
             else
                ip = (((UWord*)fp)[lr_offset]);
 
-#           if defined(VGP_ppc64_linux)
+#           if defined(VG_PLAT_USES_PPCTOC)
             /* Nasty hack to do with function replacement/wrapping on
-               ppc64-linux.  If LR points to our magic return stub,
-               then we are in a wrapped or intercepted function, in
-               which LR has been messed with.  The original LR will
-               have been pushed onto the thread's hidden REDIR stack
-               one down from the top (top element is the saved R2) and
-               so we should restore the value from there instead. */
-            if (i == 1 
-                && ip == (Addr)&VG_(ppc64_linux_magic_redirect_return_stub)
+               ppc64-linux/ppc64-aix/ppc32-aix.  If LR points to our
+               magic return stub, then we are in a wrapped or
+               intercepted function, in which LR has been messed with.
+               The original LR will have been pushed onto the thread's
+               hidden REDIR stack one down from the top (top element
+               is the saved R2) and so we should restore the value
+               from there instead.  Since nested redirections can and
+               do happen, we keep track of the number of nested LRs
+               used by the unwinding so far with 'redirs_used'. */
+            if (ip == (Addr)&VG_(ppctoc_magic_redirect_return_stub)
                 && VG_(is_valid_tid)(tid_if_known)) {
-               Long hsp = VG_(threads)[tid_if_known].arch.vex.guest_REDIR_SP;
-               if (hsp >= 1 && hsp < VEX_GUEST_PPC64_REDIR_STACK_SIZE)
+               Word hsp = VG_(threads)[tid_if_known]
+                             .arch.vex.guest_REDIR_SP;
+               hsp -= 2 * redirs_used;
+               redirs_used ++;
+               if (hsp >= 1 && hsp < redir_stack_size)
                   ip = VG_(threads)[tid_if_known]
                           .arch.vex.guest_REDIR_STACK[hsp-1];
             }
 #           endif
 
+            if (0 == ip || 1 == ip) break;
             fp = (((UWord*)fp)[0]);
-            ips[i++] = ip;
+            if (sps) sps[i] = fp; /* NB. not sp */
+            if (fps) fps[i] = fp;
+            ips[i++] = ip - 1; /* -1: refer to calling insn, not the RA */
             if (debug)
-               VG_(printf)("     ipsF[%d]=%08p\n", i-1, ips[i-1]);
+               VG_(printf)("     ipsF[%d]=%#08lx\n", i-1, ips[i-1]);
+            ip = ip - 1; /* ip is probably dead at this point, but
+                            play safe, a la x86/amd64 above.  See
+                            extensive comments above. */
             continue;
          }
 
@@ -335,7 +455,11 @@ UInt VG_(get_StackTrace2) ( ThreadId tid_if_known,
    return n_found;
 }
 
-UInt VG_(get_StackTrace) ( ThreadId tid, StackTrace ips, UInt n_ips )
+UInt VG_(get_StackTrace) ( ThreadId tid, 
+                           /*OUT*/StackTrace ips, UInt max_n_ips,
+                           /*OUT*/StackTrace sps,
+                           /*OUT*/StackTrace fps,
+                           Word first_ip_delta )
 {
    /* thread in thread table */
    Addr ip                 = VG_(get_IP)(tid);
@@ -343,6 +467,7 @@ UInt VG_(get_StackTrace) ( ThreadId tid, StackTrace ips, UInt n_ips )
    Addr sp                 = VG_(get_SP)(tid);
    Addr lr                 = VG_(get_LR)(tid);
    Addr stack_highest_word = VG_(threads)[tid].client_stack_highest_word;
+   Addr stack_lowest_word  = 0;
 
 #  if defined(VGP_x86_linux)
    /* Nasty little hack to deal with syscalls - if libc is using its
@@ -371,15 +496,25 @@ UInt VG_(get_StackTrace) ( ThreadId tid, StackTrace ips, UInt n_ips )
    }
 #  endif
 
+   /* See if we can get a better idea of the stack limits */
+   VG_(stack_limits)(sp, &stack_lowest_word, &stack_highest_word);
+
+   /* Take into account the first_ip_delta. */
+   vg_assert( sizeof(Addr) == sizeof(Word) );
+   ip += first_ip_delta;
+
    if (0)
-      VG_(printf)("tid %d: stack_highest=%p ip=%p sp=%p fp=%p\n",
+      VG_(printf)("tid %d: stack_highest=0x%08lx ip=0x%08lx "
+                  "sp=0x%08lx fp=0x%08lx\n",
 		  tid, stack_highest_word, ip, sp, fp);
 
-   return VG_(get_StackTrace2)(tid, ips, n_ips, ip, sp, fp, lr, sp, 
-                                    stack_highest_word);
+   return VG_(get_StackTrace_wrk)(tid, ips, max_n_ips, 
+                                       sps, fps,
+                                       ip, sp, fp, lr, sp, 
+                                       stack_highest_word);
 }
 
-static void printIpDesc(UInt n, Addr ip)
+static void printIpDesc(UInt n, Addr ip, void* uu_opaque)
 {
    #define BUF_LEN   4096
    
@@ -388,9 +523,9 @@ static void printIpDesc(UInt n, Addr ip)
    VG_(describe_IP)(ip, buf, BUF_LEN);
 
    if (VG_(clo_xml)) {
-      VG_(message)(Vg_UserMsg, "    %s", buf);
+      VG_(printf_xml)("    %s\n", buf);
    } else {
-      VG_(message)(Vg_UserMsg, "   %s %s", ( n == 0 ? "at" : "by" ), buf);
+      VG_(message)(Vg_UserMsg, "   %s %s\n", ( n == 0 ? "at" : "by" ), buf);
    }
 }
 
@@ -400,59 +535,54 @@ void VG_(pp_StackTrace) ( StackTrace ips, UInt n_ips )
    vg_assert( n_ips > 0 );
 
    if (VG_(clo_xml))
-      VG_(message)(Vg_UserMsg, "  <stack>");
+      VG_(printf_xml)("  <stack>\n");
 
-   VG_(apply_StackTrace)( printIpDesc, ips, n_ips );
+   VG_(apply_StackTrace)( printIpDesc, NULL, ips, n_ips );
 
    if (VG_(clo_xml))
-      VG_(message)(Vg_UserMsg, "  </stack>");
+      VG_(printf_xml)("  </stack>\n");
 }
 
 /* Get and immediately print a StackTrace. */
-void VG_(get_and_pp_StackTrace) ( ThreadId tid, UInt n_ips )
+void VG_(get_and_pp_StackTrace) ( ThreadId tid, UInt max_n_ips )
 {
-   Addr ips[n_ips];
-   UInt n_ips_obtained = VG_(get_StackTrace)(tid, ips, n_ips);
-   VG_(pp_StackTrace)(ips, n_ips_obtained);
+   Addr ips[max_n_ips];
+   UInt n_ips
+      = VG_(get_StackTrace)(tid, ips, max_n_ips,
+                            NULL/*array to dump SP values in*/,
+                            NULL/*array to dump FP values in*/,
+                            0/*first_ip_delta*/);
+   VG_(pp_StackTrace)(ips, n_ips);
 }
 
-
-void VG_(apply_StackTrace)( void(*action)(UInt n, Addr ip),
-                            StackTrace ips, UInt n_ips )
+void VG_(apply_StackTrace)(
+        void(*action)(UInt n, Addr ip, void* opaque),
+        void* opaque,
+        StackTrace ips, UInt n_ips
+     )
 {
-   #define MYBUF_LEN 50  // only needs to be long enough for 
-                         // the names specially tested for
-
    Bool main_done = False;
-   Char mybuf[MYBUF_LEN];     // ok to stack allocate mybuf[] -- it's tiny
    Int i = 0;
 
    vg_assert(n_ips > 0);
    do {
       Addr ip = ips[i];
-      if (i > 0) 
-         ip -= VG_MIN_INSTR_SZB;   // point to calling line
 
       // Stop after the first appearance of "main" or one of the other names
       // (the appearance of which is a pretty good sign that we've gone past
       // main without seeing it, for whatever reason)
-      if ( ! VG_(clo_show_below_main)) {
-         VG_(get_fnname_nodemangle)( ip, mybuf, MYBUF_LEN );
-         mybuf[MYBUF_LEN-1] = 0; // paranoia
-         if ( VG_STREQ("main", mybuf)
-#             if defined(VGO_linux)
-              || VG_STREQ("__libc_start_main", mybuf)   // glibc glibness
-              || VG_STREQ("generic_start_main", mybuf)  // Yellow Dog doggedness
-#             endif
-            )
+      if ( ! VG_(clo_show_below_main) ) {
+         Vg_FnNameKind kind = VG_(get_fnname_kind_from_IP)(ip);
+         if (Vg_FnNameMain == kind || Vg_FnNameBelowMain == kind) {
             main_done = True;
+         }
       }
 
       // Act on the ip
-      action(i, ip);
+      action(i, ip, opaque);
 
       i++;
-   } while (i < n_ips && ips[i] != 0 && !main_done);
+   } while (i < n_ips && !main_done);
 
    #undef MYBUF_LEN
 }
