@@ -7,7 +7,7 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2000-2008 Nicholas Nethercote
+   Copyright (C) 2000-2009 Nicholas Nethercote
       njn@valgrind.org
 
    This program is free software; you can redistribute it and/or
@@ -48,6 +48,11 @@ DECL_TEMPLATE(linux, sys_mount);
 DECL_TEMPLATE(linux, sys_oldumount);
 DECL_TEMPLATE(linux, sys_umount);
 
+// POSIX, but various sub-cases differ between Linux and Darwin.
+DECL_TEMPLATE(linux, sys_fcntl);
+DECL_TEMPLATE(linux, sys_fcntl64);
+DECL_TEMPLATE(linux, sys_ioctl);
+
 DECL_TEMPLATE(linux, sys_setfsuid16);
 DECL_TEMPLATE(linux, sys_setfsuid);
 DECL_TEMPLATE(linux, sys_setfsgid16);
@@ -80,12 +85,15 @@ DECL_TEMPLATE(linux, sys_pselect6);
 DECL_TEMPLATE(linux, sys_ppoll);
 
 DECL_TEMPLATE(linux, sys_epoll_create);
+DECL_TEMPLATE(linux, sys_epoll_create1);
 DECL_TEMPLATE(linux, sys_epoll_ctl);
 DECL_TEMPLATE(linux, sys_epoll_wait);
 DECL_TEMPLATE(linux, sys_epoll_pwait);
 DECL_TEMPLATE(linux, sys_eventfd);
 DECL_TEMPLATE(linux, sys_eventfd2);
 
+DECL_TEMPLATE(linux, sys_fallocate);
+
 DECL_TEMPLATE(linux, sys_gettid);
 DECL_TEMPLATE(linux, sys_set_tid_address);
 DECL_TEMPLATE(linux, sys_tkill);
@@ -252,11 +260,13 @@ DECL_TEMPLATE(linux, sys_lookup_dcookie);        // (*/32/64) L
 #define UW  UWord
 #define SR  SysRes
 
-extern void   ML_(linux_PRE_sys_msgsnd)  ( TId, UW, UW, UW, UW );
-extern void   ML_(linux_PRE_sys_msgrcv)  ( TId, UW, UW, UW, UW, UW );
-extern void   ML_(linux_POST_sys_msgrcv) ( TId, UW, UW, UW, UW, UW, UW );
-extern void   ML_(linux_PRE_sys_msgctl)  ( TId, UW, UW, UW );
-extern void   ML_(linux_POST_sys_msgctl) ( TId, UW, UW, UW, UW );
+extern void ML_(linux_PRE_sys_msgsnd)      ( TId, UW, UW, UW, UW );
+extern void ML_(linux_PRE_sys_msgrcv)      ( TId, UW, UW, UW, UW, UW );
+extern void ML_(linux_POST_sys_msgrcv)     ( TId, UW, UW, UW, UW, UW, UW );
+extern void ML_(linux_PRE_sys_msgctl)      ( TId, UW, UW, UW );
+extern void ML_(linux_POST_sys_msgctl)     ( TId, UW, UW, UW, UW );
+extern void ML_(linux_PRE_sys_getsockopt)  ( TId, UW, UW, UW, UW, UW );
+extern void ML_(linux_POST_sys_getsockopt) ( TId, SR, UW, UW, UW, UW, UW );
 
 #undef TId
 #undef UW
