@@ -7,7 +7,7 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2000-2006 Julian Seward
+   Copyright (C) 2000-2009 Julian Seward
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -42,7 +42,7 @@
 
       CORE      for the core's general use.
       TOOL      for the tool to use (and the only one it uses).
-      SYMTAB    for Valgrind's symbol table storage.
+      DINFO     for debug info (symbols, line #s, CFI, etc) storage.
       CLIENT    for the client's mallocs/frees, if the tool replaces glibc's
                     malloc() et al -- redzone size is chosen by the tool.
       DEMANGLE  for the C++ demangler.
@@ -59,7 +59,7 @@ typedef Int ArenaId;
 
 #define VG_AR_CORE         0
 #define VG_AR_TOOL         1
-#define VG_AR_SYMTAB       2
+#define VG_AR_DINFO        2
 #define VG_AR_CLIENT       3
 #define VG_AR_DEMANGLE     4
 #define VG_AR_EXECTXT      5
@@ -69,7 +69,21 @@ typedef Int ArenaId;
 // This is both the minimum payload size of a malloc'd block, and its
 // minimum alignment.  Must be a power of 2 greater than 4, and should be
 // greater than 8.
-#define VG_MIN_MALLOC_SZB        8
+#if   defined(VGP_x86_linux)   || \
+      defined(VGP_ppc32_linux) || \
+      defined(VGP_ppc32_aix5)
+#  define VG_MIN_MALLOC_SZB        8
+// Nb: We always use 16 bytes for Darwin, even on 32-bits, so it can be used
+// for any AltiVec- or SSE-related type.  This matches the Darwin libc.
+#elif defined(VGP_amd64_linux) || \
+      defined(VGP_ppc64_linux) || \
+      defined(VGP_ppc64_aix5)  || \
+      defined(VGP_x86_darwin)  || \
+      defined(VGP_amd64_darwin)
+#  define VG_MIN_MALLOC_SZB       16
+#else
+#  error Unknown platform
+#endif
 
 /* This struct definition MUST match the system one. */
 /* SVID2/XPG mallinfo structure */
@@ -86,16 +100,18 @@ struct vg_mallinfo {
    int keepcost; /* top-most, releasable (via malloc_trim) space */
 };
 
-extern void* VG_(arena_malloc)  ( ArenaId arena, SizeT nbytes );
+extern void* VG_(arena_malloc)  ( ArenaId arena, HChar* cc, SizeT nbytes );
 extern void  VG_(arena_free)    ( ArenaId arena, void* ptr );
-extern void* VG_(arena_calloc)  ( ArenaId arena, 
+extern void* VG_(arena_calloc)  ( ArenaId arena, HChar* cc,
                                   SizeT nmemb, SizeT bytes_per_memb );
-extern void* VG_(arena_realloc) ( ArenaId arena, void* ptr, SizeT size );
-extern void* VG_(arena_memalign)( ArenaId aid, SizeT req_alignB, 
-                                               SizeT req_pszB );
-extern Char* VG_(arena_strdup)  ( ArenaId aid, const Char* s);
+extern void* VG_(arena_realloc) ( ArenaId arena, HChar* cc,
+                                  void* ptr, SizeT size );
+extern void* VG_(arena_memalign)( ArenaId aid, HChar* cc,
+                                  SizeT req_alignB, SizeT req_pszB );
+extern Char* VG_(arena_strdup)  ( ArenaId aid, HChar* cc, 
+                                  const Char* s);
 
-extern SizeT VG_(arena_payload_szB) ( ThreadId tid, ArenaId aid, void* payload );
+extern SizeT VG_(arena_malloc_usable_size) ( ArenaId aid, void* payload );
 
 extern void  VG_(mallinfo) ( ThreadId tid, struct vg_mallinfo* mi );
 
@@ -103,6 +119,8 @@ extern void  VG_(sanity_check_malloc_all) ( void );
 
 extern void  VG_(print_all_arena_stats) ( void );
 
+extern void  VG_(print_arena_cc_analysis) ( void );
+
 #endif   // __PUB_CORE_MALLOCFREE_H
 
 /*--------------------------------------------------------------------*/
