@@ -7,7 +7,7 @@
    This file is part of Lackey, an example Valgrind tool that does
    some simple program measurement and tracing.
 
-   Copyright (C) 2002-2008 Nicholas Nethercote
+   Copyright (C) 2002-2009 Nicholas Nethercote
       njn@valgrind.org
 
    This program is free software; you can redistribute it and/or
@@ -132,7 +132,7 @@
 //   want to analyse locality of memory accesses -- but is not good if
 //   absolute addresses are important.
 //
-// Despite all these warnings, Dullard's results should be good enough for a
+// Despite all these warnings, Lackey's results should be good enough for a
 // wide range of purposes.  For example, Cachegrind shares all the above
 // shortcomings and it is still useful.
 //
@@ -192,15 +192,15 @@ static Bool clo_trace_sbs       = False;
 /* The name of the function of which the number of calls (under
  * --basic-counts=yes) is to be counted, with default. Override with command
  * line option --fnname. */
-static Char* clo_fnname = "_dl_runtime_resolve";
+static Char* clo_fnname = "main";
 
 static Bool lk_process_cmd_line_option(Char* arg)
 {
-   VG_STR_CLO(arg, "--fnname", clo_fnname)
-   else VG_BOOL_CLO(arg, "--basic-counts",      clo_basic_counts)
-   else VG_BOOL_CLO(arg, "--detailed-counts",   clo_detailed_counts)
-   else VG_BOOL_CLO(arg, "--trace-mem",         clo_trace_mem)
-   else VG_BOOL_CLO(arg, "--trace-superblocks", clo_trace_sbs)
+   if VG_STR_CLO(arg, "--fnname", clo_fnname) {}
+   else if VG_BOOL_CLO(arg, "--basic-counts",      clo_basic_counts) {}
+   else if VG_BOOL_CLO(arg, "--detailed-counts",   clo_detailed_counts) {}
+   else if VG_BOOL_CLO(arg, "--trace-mem",         clo_trace_mem) {}
+   else if VG_BOOL_CLO(arg, "--trace-superblocks", clo_trace_sbs) {}
    else
       return False;
    
@@ -212,12 +212,12 @@ static Bool lk_process_cmd_line_option(Char* arg)
 static void lk_print_usage(void)
 {  
    VG_(printf)(
-"    --basic-counts=no|yes     count instructions, jumps, etc. [no]\n"
+"    --basic-counts=no|yes     count instructions, jumps, etc. [yes]\n"
 "    --detailed-counts=no|yes  count loads, stores and alu ops [no]\n"
 "    --trace-mem=no|yes        trace all loads and stores [no]\n"
 "    --trace-superblocks=no|yes  trace all superblock entries [no]\n"
 "    --fnname=<name>           count calls to <name> (only used if\n"
-"                              --basic-count=yes)  [_dl_runtime_resolve]\n"
+"                              --basic-count=yes)  [main]\n"
    );
 }
 
@@ -368,17 +368,14 @@ static void instrument_detail(IRSB* sb, Op op, IRType type)
 static void print_details ( void )
 {
    Int typeIx;
-   VG_(message)(Vg_UserMsg,
-                "   Type        Loads       Stores       AluOps");
-   VG_(message)(Vg_UserMsg,
-                "   -------------------------------------------");
+   VG_(umsg)("   Type        Loads       Stores       AluOps\n");
+   VG_(umsg)("   -------------------------------------------\n");
    for (typeIx = 0; typeIx < N_TYPES; typeIx++) {
-      VG_(message)(Vg_UserMsg,
-                   "   %4s %'12llu %'12llu %'12llu",
-                   nameOfTypeIndex( typeIx ),
-                   detailCounts[OpLoad ][typeIx],
-                   detailCounts[OpStore][typeIx],
-                   detailCounts[OpAlu  ][typeIx]
+      VG_(umsg)("   %4s %'12llu %'12llu %'12llu\n",
+                nameOfTypeIndex( typeIx ),
+                detailCounts[OpLoad ][typeIx],
+                detailCounts[OpStore][typeIx],
+                detailCounts[OpAlu  ][typeIx]
       );
    }
 }
@@ -787,6 +784,27 @@ IRSB* lk_instrument ( VgCallbackClosure* closure,
             break;
          }
 
+         case Ist_CAS: {
+            /* We treat it as a read and a write of the location.  I
+               think that is the same behaviour as it was before IRCAS
+               was introduced, since prior to that point, the Vex
+               front ends would translate a lock-prefixed instruction
+               into a (normal) read followed by a (normal) write. */
+            if (clo_trace_mem) {
+               Int    dataSize;
+               IRCAS* cas = st->Ist.CAS.details;
+               tl_assert(cas->addr != NULL);
+               tl_assert(cas->dataLo != NULL);
+               dataSize = sizeofIRType(typeOfIRExpr(tyenv, cas->dataLo));
+               if (cas->dataHi != NULL)
+                  dataSize *= 2; /* since it's a doubleword-CAS */
+               addEvent_Dr( sbOut, cas->addr, dataSize );
+               addEvent_Dw( sbOut, cas->addr, dataSize );
+            }
+            addStmtToIRSB( sbOut, st );
+            break;
+         }
+
          case Ist_Exit:
             if (clo_basic_counts) {
                // The condition of a branch was inverted by VEX if a taken
@@ -865,45 +883,45 @@ static void lk_fini(Int exitcode)
       ULong total_Jccs = n_Jccs + n_IJccs;
       ULong taken_Jccs = (n_Jccs - n_Jccs_untaken) + n_IJccs_untaken;
 
-      VG_(message)(Vg_UserMsg,
-         "Counted %'llu calls to %s()", n_func_calls, clo_fnname);
+      VG_(umsg)("Counted %'llu call%s to %s()\n",
+                n_func_calls, ( n_func_calls==1 ? "" : "s" ), clo_fnname);
 
-      VG_(message)(Vg_UserMsg, "");
-      VG_(message)(Vg_UserMsg, "Jccs:");
-      VG_(message)(Vg_UserMsg, "  total:         %'llu", total_Jccs);
+      VG_(umsg)("\n");
+      VG_(umsg)("Jccs:\n");
+      VG_(umsg)("  total:         %'llu\n", total_Jccs);
       VG_(percentify)(taken_Jccs, (total_Jccs ? total_Jccs : 1),
          percentify_decs, percentify_size, percentify_buf);
-      VG_(message)(Vg_UserMsg, "  taken:         %'llu (%s)",
+      VG_(umsg)("  taken:         %'llu (%s)\n",
          taken_Jccs, percentify_buf);
       
-      VG_(message)(Vg_UserMsg, "");
-      VG_(message)(Vg_UserMsg, "Executed:");
-      VG_(message)(Vg_UserMsg, "  SBs entered:   %'llu", n_SBs_entered);
-      VG_(message)(Vg_UserMsg, "  SBs completed: %'llu", n_SBs_completed);
-      VG_(message)(Vg_UserMsg, "  guest instrs:  %'llu", n_guest_instrs);
-      VG_(message)(Vg_UserMsg, "  IRStmts:       %'llu", n_IRStmts);
+      VG_(umsg)("\n");
+      VG_(umsg)("Executed:\n");
+      VG_(umsg)("  SBs entered:   %'llu\n", n_SBs_entered);
+      VG_(umsg)("  SBs completed: %'llu\n", n_SBs_completed);
+      VG_(umsg)("  guest instrs:  %'llu\n", n_guest_instrs);
+      VG_(umsg)("  IRStmts:       %'llu\n", n_IRStmts);
       
-      VG_(message)(Vg_UserMsg, "");
-      VG_(message)(Vg_UserMsg, "Ratios:");
+      VG_(umsg)("\n");
+      VG_(umsg)("Ratios:\n");
       tl_assert(n_SBs_entered); // Paranoia time.
-      VG_(message)(Vg_UserMsg, "  guest instrs : SB entered  = %3llu : 10",
+      VG_(umsg)("  guest instrs : SB entered  = %'llu : 10\n",
          10 * n_guest_instrs / n_SBs_entered);
-      VG_(message)(Vg_UserMsg, "       IRStmts : SB entered  = %3llu : 10",
+      VG_(umsg)("       IRStmts : SB entered  = %'llu : 10\n",
          10 * n_IRStmts / n_SBs_entered);
       tl_assert(n_guest_instrs); // Paranoia time.
-      VG_(message)(Vg_UserMsg, "       IRStmts : guest instr = %3llu : 10",
+      VG_(umsg)("       IRStmts : guest instr = %'llu : 10\n",
          10 * n_IRStmts / n_guest_instrs);
    }
 
    if (clo_detailed_counts) {
-      VG_(message)(Vg_UserMsg, "");
-      VG_(message)(Vg_UserMsg, "IR-level counts by type:");
+      VG_(umsg)("\n");
+      VG_(umsg)("IR-level counts by type:\n");
       print_details();
    }
 
    if (clo_basic_counts) {
-      VG_(message)(Vg_UserMsg, "");
-      VG_(message)(Vg_UserMsg, "Exit code:       %d", exitcode);
+      VG_(umsg)("\n");
+      VG_(umsg)("Exit code:       %d\n", exitcode);
    }
 }
 
@@ -913,7 +931,7 @@ static void lk_pre_clo_init(void)
    VG_(details_version)         (NULL);
    VG_(details_description)     ("an example Valgrind tool");
    VG_(details_copyright_author)(
-      "Copyright (C) 2002-2008, and GNU GPL'd, by Nicholas Nethercote.");
+      "Copyright (C) 2002-2009, and GNU GPL'd, by Nicholas Nethercote.");
    VG_(details_bug_reports_to)  (VG_BUGS_TO);
    VG_(details_avg_translation_sizeB) ( 200 );
 
